/*****************************************************************************/
/* Copyright (C) 1993-2002 OSS Nokalva, Inc.  All rights reserved.                */
/*****************************************************************************/

/* THIS FILE IS PROPRIETARY MATERIAL OF OPEN SYSTEMS SOLUTIONS, INC.
 * AND MAY BE USED ONLY BY DIRECT LICENSEES OF OPEN SYSTEMS SOLUTIONS, INC.
 * THIS FILE MAY NOT BE DISTRIBUTED. */

/*****************************************************************************/
/* FILE: @(#)osstest.c	9.1  00/10/30		*/
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "asn1code.h"


#ifdef __NETWARE_386__
extern int ossInitNetWare(OssGlobal *, void *, char *);
#endif /* __NETWARE_386__ */

enum what {encoder, decoder, trace};

static int encodeDecodeAndPrint(OssGlobal *, int, void *, long);
static char *getMessage(OssGlobal *, char *, enum what, long);
#define MSG_SIZE 100

/*
 * Function:            ossTest() is used for testing when the -test/-wtest
 *                      option is specified on the ASN.1 compiler command line.
 *                      It illustrates use of various OSS functions available
 *                      to you.  The ASN.1 compiler generates a call to
 *                      ossTest() for each ASN.1 value reference that is
 *                      defined using the type that is a PDU if the option
 *                      -test or -wtest is specified on the command line.  You
 *                      may also choose to call ossTest() directly.
 *
 *                      ossTest() prints its input in a simple form that
 *                      closely reflects the value notation, then encodes
 *                      the input using definite length BER encoding, decodes
 *                      the encoded data, prints the decoded data in the
 *                      simple form that reflects the value notation, copies
 *                      the decoded PDU and compares the copied value to the
 *                      original value.  This is repeated using the BER
 *                      indefinite length, PER ALIGNED variant and PER
 *                      UNALIGNED variant.
 *
 * Functions exercised: The functions whose usage is illustrated are:
 *                      ossSetEncodingRules,
 *                      ossGetEncodingRules,
 *                      ossEncode,
 *                      ossDecode,
 *                      ossPrint,
 *                      ossSetEncodingFlags,
 *                      ossSetDecodingFlags,
 *                      ossGetDecodingLength,
 *                      ossGetErrMsg,
 *                      ossPrintHex,
 *                      ossPrintPDU*,
 *                      ossCpyValue**,
 *                      ossCmpValue**,
 *                      ossFreeBuf,
 *                      ossCheckConstraints**
 *
 *                      In the above list of functions "*" indicates a
 *                      function that does nothing in the time-optimized
 *                      encoder/decoder (TOED), and "**" indicates a function
 *                      that returns a code of UNIMPLEMENTED in the TOED.
 *
 * Caveats:             To exercise all the above functions fully the ASN.1
 *                      compiler must have been invoked with the options
 *                      -ber -per.
 *
 * Parameters:
 *
 *   "world"        ... pointer to OSS's global data structure
 *   "pduNum"       ... PDU number; indicates the type of the value in "data"
 *   "data"         ... input data to encode
 *
 * Return value:        value returned by ossEncode(), ossDecode(), etc.
 */
int DLL_ENTRY_FDEF ossTest(struct ossGlobal *world, int pduNum, void *data)
{

    int		rc, error = 0;
    long	run = 0; /* count of how many times ossTest() was entered */

	/*
	 * The folowing statement is needed for counting the encoder/
	 * decoder runs for each PDU.  The variable "run" may be a global
	 * one if it is "visible" only from one process at a time, but
	 * this is not the case for some environments, such as Microsoft
	 * Windows.  This use of world->userVar is just for the convenience
	 * of ossTest(); you are free to use it in whatever way you see fit.
	 *
	 * WARNING: If you did not specify -test or -wtest on the ASN.1
	 * compiler command line, you should define a variable of type
	 * long on the stack and set the pointer world->userVar to point
	 * to it, just as it is done in the ASN.1-compiler-generated control
	 * table or code file when -test or -wtest is specified.  If "run"
	 * is defined as a global variable, then you may increment it
	 * directly without the use of ossGlobal pointer world->userVar.
	 */
    if (world->userVar)
	run = ++*(long *)world->userVar;


	/* Select PER ALIGNED encoding */

    ossSetEncodingRules(world, OSS_PER_ALIGNED);

	/* Select trace information */

    ossSetEncodingFlags(world, DEBUGPDU|AUTOMATIC_ENCDEC);
    ossSetDecodingFlags(world, DEBUGPDU|AUTOMATIC_ENCDEC);

	/* Encode using PER ALIGNED, then decode and print */

    if (rc = encodeDecodeAndPrint(world, pduNum, data, run))
	error = rc;

	/* Select PER UNALIGNED encoding */

    ossSetEncodingRules(world, OSS_PER_UNALIGNED);
	
	/* Select trace information */

    ossSetEncodingFlags(world, DEBUGPDU|AUTOMATIC_ENCDEC);
    ossSetDecodingFlags(world, DEBUGPDU|AUTOMATIC_ENCDEC);

	/* Encode using PER UNALIGNED, then decode and print */

    if (rc = encodeDecodeAndPrint(world, pduNum, data, run))
	error = rc;

#if defined(_DLL) || defined(_WINDOWS) || (defined(OS2_DLL) && defined(__WIN32__))
    if (world->api == OSS_SPARTAN && world->ft.hWnd) {
	if (error) 
	    MessageBox(world->ft.hWnd,
		"Error occurred.  No tracing info is available in the Spartan version.",
						"Error", MB_OK|MB_ICONSTOP);
	else
	    MessageBox(world->ft.hWnd,
		"All values are successfully encoded and decoded.  No tracing info is available in the Spartan version.",
					"Info", MB_OK|MB_ICONINFORMATION);
    }
    else
	if (world->api == OSS_SPARTAN && !world->ft.hWnd)
	    ossPrint(world,
			"No tracing info available in the Spartan version.\n");
#elif defined(__OS2__)
	if (world->api == OSS_SPARTAN)
	    ossPrint(world, "No tracing info available in the Spartan version.\n");
#endif /* _DLL || _WINDOWS */

    return error;
}


/*
 * Function:            encodeDecodeAndPrint() is used for encoding, decoding,
 *                      and printing a PDU.
 *
 * Parameters:
 *
 *   "world"        ... pointer to OSS's global data structure
 *   "pduNum"       ... PDU number
 *   "data"         ... input data
 *   "run"          ... count of how many times ossTest() has been entered
 *
 * Return value:        an integer returned by ossEncode(), ossDecode(), etc.
 */
static int encodeDecodeAndPrint(OssGlobal *world, int pduNum, void *data,
				long run)
{

    char      *buf;              /* Encoder output buffer address to free */
    OssBuf     encodedData;      /* OSS encoder/decoder output/input */
    void      *decodedData;      /* OSS decoder output data */
    void      *copiedValue = NULL; /* Copied PDU */
    int        rc, error = 0;    /* Return & error codes */
    char       msgBuf[MSG_SIZE]; /* Message buffer */

    encodedData.value = NULL;
    encodedData.length = 0;
    ossPrint(world, getMessage(world, msgBuf, encoder, run));
			/*
			 * Print the data pointed to by "data"
			 */
    ossPrint(world, "\n\tUnencoded PDU...\n\n");
    if (rc = ossPrintPDU(world, pduNum, data)) {
	ossPrint(world, "OssPrintPDU failed with return code = %d\n", rc);
	error = rc;
    }
	/*
	 * Check constraints
	 */
    ossPrint(world, "\n\tChecking constraints...\n\n");
    if (rc = ossCheckConstraints(world, pduNum, data)) {
	ossPrint(world, "OssCheckConstraints failed with return code = %d\n", rc);
	error = rc;
    }
    else
	ossPrint(world, "Constraints checking succeeded\n");
			/*
			 * Encode the data pointed to by "data"
			 */
    ossPrint(world, getMessage(world, msgBuf, trace, run));
    if (rc = ossEncode(world, pduNum, data, &encodedData)) {
	ossPrint(world, "Encoding failed with return code = %d\n%s", rc, ossGetErrMsg(world));
	error = rc;
    }
    else {
	ossPrint(world, "\nPDU successfully encoded, in %ld bytes:\n", encodedData.length);
			/*
			 * Print in hex the encoded data
			 */
	ossPrintHex(world, (char *)encodedData.value, encodedData.length);
	buf  = (char *)encodedData.value;
			/*
			 * Setting decodedData to NULL means that the
			 * decoder buffer is not provided
			 */
	decodedData = NULL;


	ossPrint(world, getMessage(world, msgBuf, decoder, run));

			/*
			 * Decode the PDU that was just encoded 
			 */
	ossPrint(world, "\n\tTracing Information from Decoder...\n\n");
	if (rc = ossDecode(world, &pduNum, &encodedData, &decodedData)) {
	    ossPrint(world, "Decoding failed with return code = %d\n%s", rc, ossGetErrMsg(world));
	    error = rc;
	}
	else {
	    ossPrint(world, "\nPDU successfully decoded, in %ld bytes\n",
			    ossGetDecodingLength(world));
	    ossPrint(world, "\n\tDecoded PDU...\n\n");
			/*
			 * Print the PDU that was just decoded
			 */
	    if (rc = ossPrintPDU(world, pduNum, decodedData)) {
		ossPrint(world, "OssPrintPDU failed with return code = %d\n", rc);
		error = rc;
	    }
			/*
			 * Copy the decoded PDU
			 */
	    ossPrint(world, "\nCopying the Decoded PDU...\n");
	    rc = ossCpyValue(world, pduNum, decodedData, &copiedValue);
	    if (rc) {
		ossPrint(world, "OssCpyValue failed with return code = %d\n", rc);
		error = rc;
	    }
	    else {
			/*
			 * Compare the decoded and copied PDUs
			 */
		ossPrint(world, "Comparing the Original and Copied PDUs...\n");
		rc = ossCmpValue(world, pduNum, data, copiedValue);
		if (rc) {
		    ossPrint(world, "OssCmpValue failed with return code = %d\n", rc);
		    error = rc;
		}
		else
		    ossPrint(world, "Value comparison succeeded\n");
			/*
			 * Free the copied PDU
			 */
		if (rc = ossFreePDU(world, pduNum, copiedValue)) {
		    ossPrint(world, "OssFreePDU failed with return code = %d\n", rc);
		    error = rc;
		}
	    }
			/*
			 * Free the decoder's output buffer
			 */
	    if (rc = ossFreePDU(world, pduNum, decodedData)) {
		ossPrint(world, "OssFreePDU failed with return code = %d\n", rc);
		error = rc;
	    }
	}
			/*
			 * Free the encoder's output buffer
			 */
	ossFreeBuf(world, buf);
    }
    return error;
}


/*
 * Function:            getMessage() formats and returns a message
 *
 * Parameters:
 *
 *   "world"        ... pointer to OSS's global data structure
 *   "buf"          ... message buffer
 *   "call"         ... enumerated indicating what is going to be called
 *   "run"          ... count of how many times ossTest() has been entered
 *
 * Return value:        pointer to the message buffer
 */
static char *getMessage(OssGlobal *world, char *buf, enum what call, long run)
{

    unsigned int flags = ossGetEncodingFlags(world);
    char	*newline = (run == 1 && call == encoder && flags & DEFINITE)?
			"": "\n";



    switch (ossGetEncodingRules(world)) {
	case OSS_DER:
	    switch (call) {
		case encoder:
		    sprintf(buf,
	"\n---------- Run %ld of the DER Encoder ----------\n", run);
		    break;
		case decoder:
		    sprintf(buf,
	"\n========== Run %ld of the DER Decoder ==========\n", run);
		    break;
		case trace:
		    sprintf(buf,
			"\n\tTracing Information from the DER Encoder...\n\n");
	    }
	    break;
	case OSS_BER:
	    switch (call) {
		case encoder:
		    sprintf(buf, (DEFINITE & ossGetEncodingFlags(world))?
	"%s---------- Run %ld of the BER DEFINITE-Length Encoder ----------\n":
	"%s---------- Run %ld of the BER INDEFINITE-Length Encoder ----------\n",
		    newline, run);
		    break;
		case decoder:
		    sprintf(buf, (DEFINITE & ossGetEncodingFlags(world))?
	"\n========== Run %ld of the BER Decoder for DEFINITE PDU ==========\n":
	"\n========== Run %ld of the BER Decoder for INDEFINITE PDU ==========\n", run);
		    break;
		case trace:
		    sprintf(buf, (DEFINITE & ossGetEncodingFlags(world))?
			"\n\tTracing Information from the BER DEFINITE-Length Encoder...\n\n":
			"\n\tTracing Information from the BER INDEFINITE-Length Encoder...\n\n");
	    }
	    break;
	case OSS_PER_ALIGNED:
	case OSS_PER_UNALIGNED:
	    switch (call) {
		case encoder:
		    sprintf(buf, (OSS_PER_ALIGNED == ossGetEncodingRules(world))?
	"\n---------- Run %ld of the PER ALIGNED Encoder ----------\n":
	"\n---------- Run %ld of the PER UNALIGNED Encoder ----------\n", run);
		    break;
		case decoder:
		    sprintf(buf, (OSS_PER_ALIGNED == ossGetEncodingRules(world))?
	"\n========== Run %ld of the PER ALIGNED Decoder ==========\n":
	"\n========== Run %ld of the PER UNALIGNED Decoder ==========\n", run);
		    break;
		case trace:
		    sprintf(buf, (OSS_PER_ALIGNED == ossGetEncodingRules(world))?
			"\n\tTracing Information from the PER ALIGNED Encoder...\n\n":
			"\n\tTracing Information from the PER UNALIGNED Encoder...\n\n");
	    }
    }
    return buf;
}


#ifdef __NETWARE_386__
int ossInitNetWare(OssGlobal *world, void *ctl_tbl, char *name)
{

    int		rc;
    char	path[PATH_MAX];
    double	d = 0.0;	/* Forces the Watcom compiler reference
				 * and later link the floating point format. */


    if (!(rc = ossinit(world, ctl_tbl))) {
	world->asn1prnt = fprintf;
	strcpy(path, name);
	strcpy(1 + strchr(path, '.'), "out");
	world->asn1out = fopen(path, "wt");
	if (!world->asn1out) {
	    fprintf(stdout, "Can't open trace file %s.\n", path);
	    return CANT_OPEN_TRACE_FILE;
	}
    }
    else
	/*
	 * Passing the dummy "d" below forces the Watcom compiler reference
	 * and later link the floating point format.
	 */
	fprintf(stdout, "Initialization failed with an error code of %d.\n", rc, d);
    return rc;
}
#endif /* __NETWARE_386__ */

#define NUMBER_OF_OCTETS_TO_WRITE_PER_LINE      32
/****************************************************************************
 * Function:            ossPrintEncodedMessage() is used to print encoded   *
 *                      data in hex format. The sequence of identical bytes *
 *                      is printed in the short form:                       *
 *                      "<byte_value>...(rep=<length_of_sequence>)"         *
 *                                                                          *
 * Parameters:                                                              *
 *   "world"        ... points to the ossGlobal structure that contains     *
 *                      an information about the OSS run-time environment   *
 *   "encoded_data" ... pointer to the data buffer to be printed out        *
 *   "length"       ... buffer length                                       *
 *                                                                          *
 * Return value:        NONE                                                *
 ****************************************************************************/
static void     ossPrintEncodedMessage(OssGlobal *world,
				unsigned char *encoded_data,
				long length)
{
    long            bytes_printed = 0;
    int             num_of_byte = 0;
    int             repeated = 0;
    unsigned char   saved_byte;

    if (length < 1024) {
	ossPrintHex(world, (char *)encoded_data, length);
	return;
    }

    while (bytes_printed < length || repeated) {
	if (repeated) {
	    /* if we have sequence of identical bytes which was not *
	     * printed in short form, we print each byte separately */
	    ossPrint(world, "%c", "0123456789ABCDEF"[saved_byte >> 4 & 0x0F]);
	    ossPrint(world, "%c", "0123456789ABCDEF"[saved_byte & 0x0F]);
	    --repeated;
	    ++num_of_byte;
	} else {
	    ossPrint(world, "%c",
		"0123456789ABCDEF"[encoded_data[bytes_printed] >> 4 & 0x0F]);
	    ossPrint(world, "%c",
		"0123456789ABCDEF"[encoded_data[bytes_printed] & 0x0F]);
	    ++num_of_byte;
	    saved_byte = encoded_data[bytes_printed];
	    ++bytes_printed;

	    /* count length of a sequence of identical bytes */
	    while (saved_byte == encoded_data[bytes_printed] &&
		    bytes_printed < length) {
		++repeated;
		++bytes_printed;
	    }
 
	    if (repeated > 6) {
		char             buffer[20];
		unsigned int     output_position;

		/* if sequence found is large enough, *
		 * it will be printed it shortly      */
		sprintf(buffer, "...(length=%d) ", ++repeated);
		ossPrint(world, buffer);
		output_position = num_of_byte * 2 + (num_of_byte - 1) / 4
				    + strlen(buffer);
		/* move to the output position of next group of bytes */
		while ((output_position + 1) % 9) {
		    ossPrint(world, "%c",  ' ');
		    ++output_position;
		}
		/* calculate the number of byte at this position */
		num_of_byte = (output_position - output_position / 9) / 2;
		repeated = 0;
	    }
	}
 
	if (num_of_byte >= NUMBER_OF_OCTETS_TO_WRITE_PER_LINE) {
	    ossPrint(world, "%c",  '\n');
	    num_of_byte = 0;
	} else if (!(num_of_byte & 3))
	    ossPrint(world, "%c",  ' ');
    }
	if (num_of_byte != 0)
		ossPrint(world, "%c",  '\n');
}

/****************************************************************************
 * Function:            ossTestEncoding() is the function used to check an  *
 *                      encoded message which passed as input argument.     *
 *                                                                          *
 * Parameters:                                                              *
 *   "world"        ... points to the ossGlobal structure that contains     *
 *                      an information about the OSS run-time environment   *
 *   "pduNum"       ... number of the PDU that should be tested.            *
 *   "encodedData"  ... encoded message                                     *
 *                                                                          *
 * Return value:        value returned by ossEncode(), ossDecode(), etc.    *
 ****************************************************************************/
int DLL_ENTRY_FDEF ossTestEncoding(OssGlobal *world, int pduNum,
				    OssBuf *encodedData)
{
    int             pdu_num = pduNum;
    int             ret_code;
    void            *data = NULL;
    int             rules;
    char           *rules_name;
    long            run = 1;
    unsigned long   bytes_read;
 
    switch (rules = ossGetEncodingRules(world)) {
	case OSS_PER_ALIGNED:
	    rules_name = "PER ALIGNED";
	    break;
	case OSS_PER_UNALIGNED:
	    rules_name = "PER UNALIGNED";
	    break;
	case OSS_DER:
	    rules_name = "DER";
	    break;
	default:
	    rules_name = "BER";
	    rules = OSS_BER;
	    break;
    }
    if (world->userVar)
	run = *(long *)world->userVar + 1;
    ossPrint(world,
	"\n########## Run %2ld of the %s Decoder ##########\n\n",
				    run, rules_name );
    ossPrint(world, "Encoding:\n");
    ossPrintEncodedMessage(world, encodedData->value, encodedData->length);

    ossSetEncodingRules(world, rules);
    ret_code = ossDecode(world, &pdu_num, encodedData, &data);
    if (ret_code != PDU_DECODED) {
	ossPrint(world, "\nDecoder error code: %d\n", ret_code);
	if (world->userVar)
	    ++*(long *)world->userVar;
	return ret_code;
    }
    bytes_read = ossGetBytesReadByDecoder(world);
    encodedData->length -= bytes_read;
    encodedData->value += bytes_read;

    ossPrint(world, "\n\tCall ossTest() to check a decoded PDU... \n");
    ret_code = ossTest(world, pdu_num, data);
    if (ret_code)
	ossPrint(world, "Testing failed with return code = %d\n", ret_code);

    ossFreePDU(world, pdu_num, data);
 
    return ret_code;
}

/****************************************************************************
 * Function:            ossTestMultipleEncodings() is the function used     *
 *                      to check an array of encoded messages (one by       *
 *                      another).                                           *
 *                                                                          *
 * Parameters:                                                              *
 *   "world"        ... points to the ossGlobal structure that contains     *
 *                      an information about the OSS run-time environment   *
 *   "pduNum"       ... number of the PDU that should be tested.            *
 *   "encoded_message"                                                      *
 *                  ... is array of encoded messages                        *
 *   "number_of_encoded_messages"                                           *
 *                  ... number of encoded messages                          *
 *                                                                          *
 * Return value:        number of bad encodings (0 if no errors were        *
 *                      detected).                                          *
 ****************************************************************************/
int DLL_ENTRY_FDEF ossTestMultipleEncodings(OssGlobal *world, int pduNum,
			    OssBuf *encoded_message,
			    int number_of_encoded_messages)
{
    int             step;
    int             error_counter = 0;
    int             rules = ossGetEncodingRules(world);

    for (step = 0; step < number_of_encoded_messages; step++) {
	ossSetEncodingRules(world, rules);
	if (ossTestEncoding(world, pduNum, &encoded_message[step]))
	    error_counter++;
    }

    return error_counter;
}

/****************************************************************************
 * Function:            ossTestConcatenatedEncodings() tests an encoding    *
 *                      which consist of multiple concatenated encodings    *
 *                      are placed in a single input buffer.                *
 *                                                                          *
 * Parameters:                                                              *
 *   "world"        ... points to the ossGlobal structure that contains     *
 *                      an information about the OSS run-time environment   *
 *   "pduNum"       ... number of the PDU that should be tested.            *
 *   "encoded_message"                                                      *
 *                  ... is buffer of concatenated encoded messages          *
 *                                                                          *
 * Return value:        0 if no errors were detected, otherwise the code    *
 *                      returned by the ossTestEncoding()                   *
 ****************************************************************************/
int DLL_ENTRY_FDEF ossTestConcatenatedEncodings(OssGlobal *world, int pduNum,
			    OssBuf *encoded_message)
{
    int             rules = ossGetEncodingRules(world);
    int             ret_code = 0;

    while (encoded_message->length > 0) {
	ossSetEncodingRules(world, rules);
	if ( (ret_code = ossTestEncoding(world, pduNum, encoded_message)) )
	    break;
    }

    return ret_code;
}
